#include <stdio.h>
#include <elf.h>
#include <link.h>

/*
 * With the help of:
 * grep -E "^#define DT_" /usr/include/elf.h | \
 * sed 's/^#define[[:space:]]\+\([a-zA-Z0-9_]\+\).\+/[\1] = \"\1\",/'
 */
static const char *dyn_tagnames[] = {
	[DT_NULL] = "DT_NULL",
	[DT_NEEDED] = "DT_NEEDED",
	[DT_PLTRELSZ] = "DT_PLTRELSZ",
	[DT_PLTGOT] = "DT_PLTGOT",
	[DT_HASH] = "DT_HASH",
	[DT_STRTAB] = "DT_STRTAB",
	[DT_SYMTAB] = "DT_SYMTAB",
	[DT_RELA] = "DT_RELA",
	[DT_RELASZ] = "DT_RELASZ",
	[DT_RELAENT] = "DT_RELAENT",
	[DT_STRSZ] = "DT_STRSZ",
	[DT_SYMENT] = "DT_SYMENT",
	[DT_INIT] = "DT_INIT",
	[DT_FINI] = "DT_FINI",
	[DT_SONAME] = "DT_SONAME",
	[DT_RPATH] = "DT_RPATH",
	[DT_SYMBOLIC] = "DT_SYMBOLIC",
	[DT_REL] = "DT_REL",
	[DT_RELSZ] = "DT_RELSZ",
	[DT_RELENT] = "DT_RELENT",
	[DT_PLTREL] = "DT_PLTREL",
	[DT_DEBUG] = "DT_DEBUG",
	[DT_TEXTREL] = "DT_TEXTREL",
	[DT_JMPREL] = "DT_JMPREL",
	[DT_BIND_NOW] = "DT_BIND_NOW",
	[DT_INIT_ARRAY] = "DT_INIT_ARRAY",
	[DT_FINI_ARRAY] = "DT_FINI_ARRAY",
	[DT_INIT_ARRAYSZ] = "DT_INIT_ARRAYSZ",
	[DT_FINI_ARRAYSZ] = "DT_FINI_ARRAYSZ",
	[DT_RUNPATH] = "DT_RUNPATH",
	[DT_FLAGS] = "DT_FLAGS",
	[DT_ENCODING] = "DT_ENCODING",
	[DT_PREINIT_ARRAY] = "DT_PREINIT_ARRAY",
	[DT_PREINIT_ARRAYSZ] = "DT_PREINIT_ARRAYSZ",
};




int main(int argc, char *argv[], char *envp[])
{
	ElfW(Dyn) *dyn;
	ElfW(auxv_t) *auxv;
	struct r_debug *r_debug;
	struct link_map *map;
	ElfW(Addr) entry;
	ElfW(Addr) vdso_l_map_start;
	ElfW(Addr) vdso_l_map_end;
	ElfW(Addr) vdso_l_addr;
	ElfW(Addr) vdso_l_ld;

	/* *envp = NULL marks end of envp */
	while(*envp++ != NULL);

	/* auxv->a_type = AT_NULL marks the end of auxv */
	for (auxv = (ElfW(auxv_t) *)envp; auxv->a_type != AT_NULL; auxv++) {
		int i;

		switch(auxv->a_type) {
		case AT_SYSINFO_EHDR:
			printf("AT_SYSINFO_EHDR:\t%#lx\n",
			       auxv->a_un.a_val);

			ElfW(Ehdr) *ehdr = auxv->a_un.a_val;
			ElfW(Phdr) *phdr = auxv->a_un.a_val + ehdr->e_phoff;
			printf("e_phoff=%#lx, e_phnum=%#lx\n",
			       ehdr->e_phoff, ehdr->e_phnum);

			for (i = 0; i < ehdr->e_phnum; ++i) {
				const ElfW(Phdr) *const ph = &phdr[i];
				printf("ph=%#lx, p_type=%#lx, p_vaddr=%#lx\n",
				       ph, ph->p_type, ph->p_vaddr);
				switch(ph->p_type) {
				case PT_LOAD:
					vdso_l_addr = ph->p_vaddr;
					vdso_l_map_end = ph->p_vaddr +
						ph->p_memsz;
					break;
				case PT_DYNAMIC:
					vdso_l_ld = ph->p_vaddr;
					break;
				case PT_TLS:
					/* There must be no TLS segment.  */
					abort();
					break;
				}
			}

			vdso_l_map_start = (ElfW(Addr))auxv->a_un.a_val;
			vdso_l_addr = vdso_l_map_start - vdso_l_addr;
			vdso_l_map_end += vdso_l_addr;
			vdso_l_ld = vdso_l_ld + vdso_l_addr;
			printf("vDSO: l_addr=%#lx, l_ld=%#lx, l_map_end=%#lx\n",
			      vdso_l_addr, vdso_l_ld, vdso_l_map_end);
			break;
		case AT_ENTRY:
			printf("AT_ENTRY:\t%#lx\n",
			       auxv->a_un.a_val);
			entry = auxv->a_un.a_val;
			break;
		}
	}

	for (dyn = _DYNAMIC; dyn->d_tag != DT_NULL; ++dyn)
		if (dyn->d_tag == DT_DEBUG)
			r_debug = (struct r_debug *) dyn->d_un.d_ptr;

	for(map = r_debug->r_map; map; map = map->l_next) {
		char *strings;

		printf("Link Map Entry:\n\
\tlink_map=%p\n\
\tl_addr=%#lx,\n\
\tl_name=\"%s\",\n\
\tl_ld=%#lx\n",
		       map,
		       map->l_addr,
		       map->l_name, (unsigned long)map->l_ld);

		if (map->l_addr == vdso_l_addr) {
			printf("\tvDSO detected\n");
			continue;
		}

		for (dyn = map->l_ld; dyn->d_tag != DT_NULL; ++dyn) {
			char buf[19];
			const char *d_tag;

			if (dyn->d_tag >= DT_NUM) {
				snprintf (buf, sizeof(buf), "%#x", dyn->d_tag);
				d_tag = buf;
			} else
				d_tag = dyn_tagnames[dyn->d_tag];

			printf("\t%s=%u(%#lx)\n", d_tag, dyn->d_un.d_val,
			       dyn->d_un.d_ptr);

			switch(dyn->d_tag) {
			case DT_STRTAB:
				strings = (char *) dyn->d_un.d_ptr;
				break;
			}
		}

		for (dyn = map->l_ld; dyn->d_tag != DT_NULL; ++dyn)
			if (dyn->d_tag == DT_SONAME)
				printf("DT_SONAME=%d (0x%lx)\n", dyn->d_un.d_val,
					(char *)strings+dyn->d_un.d_val);
	}
}
